# __docformat__ = "restructuredtext en"
# ******NOTICE***************
# From optimize.py module by Travis E. Oliphant
#
# You may copy and use this module as you see fit with no
# guarantee implied provided you keep this notice in all copies.
# *****END NOTICE************
#
# Modified by Paul Kienzle to support bounded minimization
"""
Downhill simplex optimizer.
"""

__all__ = ["simplex"]
__docformat__ = "restructuredtext en"
__version__ = "0.7"

import numpy as np


def wrap_function(function, bounds):
    ncalls = [0]
    if bounds is not None:
        lo, hi = [np.asarray(v) for v in bounds]

        def function_wrapper(x):
            ncalls[0] += 1
            if np.any((x < lo) | (x > hi)):
                return np.inf
            else:
                # function(x)
                return function(x)
    else:

        def function_wrapper(x):
            ncalls[0] += 1
            return function(x)

    return ncalls, function_wrapper


class Result(object):
    """
    Results from the fit.

    x : ndarray
        Best parameter set
    fx : float
        Best value
    iters : int
        Number of iterations
    calls : int
        Number of function calls
    status : boolean
        True if the fit completed successful, false if terminated early
        because of too many iterations.
    """

    def __init__(self, x, fx, iters, calls, status):
        self.x, self.fx, self.iters, self.calls = x, fx, iters, calls
        self.status = status

    def __str__(self):
        msg = "Converged" if self.status else "Aborted"
        return "%s with %g at %s after %d calls" % (msg, self.fx, self.x, self.calls)


def dont_abort():
    return False


def simplex(
    f, x0=None, bounds=None, radius=0.05, xtol=1e-4, ftol=1e-4, maxiter=None, update_handler=None, abort_test=dont_abort
):
    """
    Minimize a function using Nelder-Mead downhill simplex algorithm.

    This optimizer is also known as Amoeba (from Numerical Recipes) and
    the Nealder-Mead simplex algorithm.  This is not the simplex algorithm
    for solving constrained linear systems.

    Downhill simplex is a robust derivative free algorithm for finding
    minima.  It proceeds by choosing a set of points (the simplex) forming
    an n-dimensional triangle, and transforming that triangle so that the
    worst vertex is improved, either by stretching, shrinking or reflecting
    it about the center of the triangle.  This algorithm is not known for
    its speed, but for its simplicity and robustness, and is a good algorithm
    to start your problem with.

    *Parameters*:

        f : callable f(x,*args)
            The objective function to be minimized.
        x0 : ndarray
            Initial guess.
        bounds : (ndarray,ndarray) or None
            Bounds on the parameter values for the function.
        radius: float
            Size of the initial simplex.  For bounded parameters (those
            which have finite lower and upper bounds), radius is clipped
            to a value in (0,0.5] representing the portion of the
            range to use as the size of the initial simplex.

    *Returns*: Result (`park.simplex.Result`)

        x : ndarray
            Parameter that minimizes function.
        fx : float
            Value of function at minimum: ``fopt = func(xopt)``.
        iters : int
            Number of iterations performed.
        calls : int
            Number of function calls made.
        success : boolean
            True if fit completed successfully.

    *Other Parameters*:

        xtol : float
            Relative error in xopt acceptable for convergence.
        ftol : number
            Relative error in func(xopt) acceptable for convergence.
        maxiter : int=200*N
            Maximum number of iterations to perform.  Defaults
        update_handler : callable
            Called after each iteration, as callback(k,n,xk,fxk),
            where k is the current iteration, n is the maximum
            iteration, xk is the simplex and fxk is the value of
            the simplex vertices.  xk[0],fxk[0] is the current best.
        abort_test : callable
            Called after each iteration, as callback(), to see if
            an external process has requested stop.

    *Notes*

        Uses a Nelder-Mead simplex algorithm to find the minimum of
        function of one or more variables.

    """
    fcalls, func = wrap_function(f, bounds)
    x0 = np.asarray(x0, dtype=float).flatten()
    # print "x0",x0
    N = len(x0)
    rank = len(x0.shape)
    if not -1 < rank < 2:
        raise ValueError("Initial guess must be a scalar or rank-1 sequence.")

    if maxiter is None:
        maxiter = N * 200

    rho = 1
    chi = 2
    psi = 0.5
    sigma = 0.5

    if rank == 0:
        sim = np.zeros((N + 1,), dtype=x0.dtype)
    else:
        sim = np.zeros((N + 1, N), dtype=x0.dtype)
    fsim = np.zeros((N + 1,), float)
    sim[0] = x0
    fsim[0] = func(x0)

    # Metropolitan simplex: simplex has vertices at x0 and at
    # x0 + j*radius for each unit vector j.  Radius is a percentage
    # change from the initial value, or just the radius if the initial
    # value is 0.  For bounded problems, the radius is a percentage of
    # the bounded range in dimension j.
    val = x0 * (1 + radius)
    val[val == 0] = radius
    if bounds is not None:
        radius = np.clip(radius, 0, 0.5)
        lo, hi = [np.asarray(v) for v in bounds]

        # Keep the initial simplex inside the bounds
        x0 = np.select([x0 < lo, x0 > hi], [lo, hi], x0)
        bounded = ~np.isinf(lo) & ~np.isinf(hi)
        val[bounded] = x0[bounded] + (hi[bounded] - lo[bounded]) * radius
        val = np.select([val < lo, val > hi], [lo, hi], val)

        # If the initial point was at or beyond an upper bound, then bounds
        # projection will put x0 and x0+j*radius at the same point.  We
        # need to detect these collisions and reverse the radius step
        # direction when such collisions occur.  The only time the collision
        # can occur at the lower bound is when upper and lower bound are
        # identical.  In that case, we are already done.
        collision = val == x0
        if np.any(collision):
            reverse = x0 * (1 - radius)
            reverse[reverse == 0] = -radius
            reverse[bounded] = x0[bounded] - (hi[bounded] - lo[bounded]) * radius
            val[collision] = reverse[collision]

        # Make tolerance relative for bounded parameters
        tol = np.ones(x0.shape) * xtol
        tol[bounded] = (hi[bounded] - lo[bounded]) * xtol
        xtol = tol

    # Compute values at the simplex vertices
    for k in range(0, N):
        y = x0 + 0
        y[k] = val[k]
        sim[k + 1] = y
        fsim[k + 1] = func(y)

    # print sim
    ind = np.argsort(fsim)
    fsim = np.take(fsim, ind, 0)
    # sort so sim[0,:] has the lowest function value
    sim = np.take(sim, ind, 0)
    # print sim

    iterations = 1
    while iterations < maxiter:
        if np.all(abs(sim[1:] - sim[0]) <= xtol) and max(abs(fsim[0] - fsim[1:])) <= ftol:
            # print abs(sim[1:]-sim[0])
            break

        xbar = np.sum(sim[:-1], 0) / N
        xr = (1 + rho) * xbar - rho * sim[-1]
        # print "xbar" ,xbar,rho,sim[-1],N
        # break
        fxr = func(xr)
        doshrink = 0

        if fxr < fsim[0]:
            xe = (1 + rho * chi) * xbar - rho * chi * sim[-1]
            fxe = func(xe)

            if fxe < fxr:
                sim[-1] = xe
                fsim[-1] = fxe
            else:
                sim[-1] = xr
                fsim[-1] = fxr
        else:  # fsim[0] <= fxr
            if fxr < fsim[-2]:
                sim[-1] = xr
                fsim[-1] = fxr
            else:  # fxr >= fsim[-2]
                # Perform contraction
                if fxr < fsim[-1]:
                    xc = (1 + psi * rho) * xbar - psi * rho * sim[-1]
                    fxc = func(xc)

                    if fxc <= fxr:
                        sim[-1] = xc
                        fsim[-1] = fxc
                    else:
                        doshrink = 1
                else:
                    # Perform an inside contraction
                    xcc = (1 - psi) * xbar + psi * sim[-1]
                    fxcc = func(xcc)

                    if fxcc < fsim[-1]:
                        sim[-1] = xcc
                        fsim[-1] = fxcc
                    else:
                        doshrink = 1

                if doshrink:
                    for j in range(1, N + 1):
                        sim[j] = sim[0] + sigma * (sim[j] - sim[0])
                        fsim[j] = func(sim[j])

        ind = np.argsort(fsim)
        sim = np.take(sim, ind, 0)
        fsim = np.take(fsim, ind, 0)
        if update_handler is not None:
            update_handler(iterations, maxiter, sim, fsim)
        iterations += 1
        if abort_test():
            break  # STOPHERE

    status = 0 if iterations < maxiter else 1
    res = Result(sim[0], fsim[0], iterations, fcalls[0], status)
    res.next_start = sim[np.random.randint(N)]
    return res


def main():
    import time

    def rosen(x):  # The Rosenbrock function
        return np.sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0, axis=0)

    x0 = [0.8, 1.2, 0.7]
    print("Nelder-Mead Simplex")
    print("===================")
    start = time.time()
    x = simplex(rosen, x0)
    print(x)
    print("Time:", time.time() - start)

    x0 = [0] * 3
    print("Nelder-Mead Simplex")
    print("===================")
    print("starting at zero")
    start = time.time()
    x = simplex(rosen, x0)
    print(x)
    print("Time:", time.time() - start)

    x0 = [0.8, 1.2, 0.7]
    lo, hi = [0] * 3, [1] * 3
    print("Bounded Nelder-Mead Simplex")
    print("===========================")
    start = time.time()
    x = simplex(rosen, x0, bounds=(lo, hi))
    print(x)
    print("Time:", time.time() - start)

    x0 = [0.8, 1.2, 0.7]
    lo, hi = [0.999] * 3, [1.001] * 3
    print("Bounded Nelder-Mead Simplex")
    print("===========================")
    print("tight bounds")
    print("simplex is smaller than 1e-7 in every dimension, but you can't")
    print("see this without uncommenting the print statement simplex function")
    start = time.time()
    x = simplex(rosen, x0, bounds=(lo, hi), xtol=1e-4)
    print(x)
    print("Time:", time.time() - start)

    x0 = [0] * 3
    hi, lo = [-0.999] * 3, [-1.001] * 3
    print("Bounded Nelder-Mead Simplex")
    print("===========================")
    print("tight bounds, x0=0 outside bounds from above")
    start = time.time()
    x = simplex(lambda x: rosen(-x), x0, bounds=(lo, hi), xtol=1e-4)
    print(x)
    print("Time:", time.time() - start)

    x0 = [0.8, 1.2, 0.7]
    lo, hi = [-np.inf] * 3, [np.inf] * 3
    print("Bounded Nelder-Mead Simplex")
    print("===========================")
    print("infinite bounds")
    start = time.time()
    x = simplex(rosen, x0, bounds=(lo, hi), xtol=1e-4)
    print(x)
    print("Time:", time.time() - start)


if __name__ == "__main__":
    main()



import asyncio
from copy import deepcopy
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    Optional,
    Dict,
    List,
    NewType,
    Tuple,
    TypedDict,
    Any,
    Literal,
    Union,
    cast,
    IO,
)
from collections import deque
from dataclasses import dataclass, field, fields
import json
import shutil
import os
import tempfile
from pathlib import Path
from threading import Event
from bumps.serialize import serialize, deserialize
from bumps.util import get_libraries
import h5py
import numpy as np

from bumps.dream.state import MCMCDraw
from .logger import logger

if TYPE_CHECKING:
    import bumps, bumps.fitproblem, bumps.dream.state
    from .webserver import TopicNameType
    from .fit_thread import FitThread
    from h5py import Group, Dataset


SESSION_FILE_NAME = "session.h5"
MAX_PROBLEM_SIZE = 100 * 1024 * 1024  # 10 MB problem max size
UNCERTAINTY_DTYPE = "f"
MAX_LABEL_LENGTH = 1024
LABEL_DTYPE = f"|S{MAX_LABEL_LENGTH}"
COMPRESSION = 5
UNCERTAINTY_COMPRESSION = 5

SERIALIZERS = Literal["dataclass", "pickle", "dill"]
SERIALIZER_EXTENSIONS = {"dataclass": "json", "pickle": "pickle", "dill": "pickle"}
DEFAULT_SERIALIZER: SERIALIZERS = "dill"


@dataclass(frozen=True)
class UNDEFINED_TYPE:
    pass


UNDEFINED = UNDEFINED_TYPE()


def now_string():
    return f"{datetime.now().timestamp():.6f}"


def serialize_problem(problem: "bumps.fitproblem.FitProblem", method: SERIALIZERS):
    if method == "dataclass":
        return json.dumps(serialize(problem)).encode()
    elif method == "pickle":
        import pickle

        return pickle.dumps(problem)
    elif method == "dill":
        import dill

        return dill.dumps(problem, recurse=True)


def deserialize_problem(serialized: bytes, method: SERIALIZERS):
    if method == "dataclass":
        serialized_dict = json.loads(serialized)
        return deserialize(serialized_dict, migration=True)
    elif method == "pickle":
        import pickle

        return pickle.loads(serialized)
    elif method == "dill":
        import dill

        return dill.loads(serialized)


def write_bytes_data(group: "Group", name: str, data: bytes):
    saved_data = [data] if data is not None else []
    return group.create_dataset(name, data=np.void(saved_data), compression=COMPRESSION)


def read_bytes_data(group: "Group", name: str):
    if not name in group:
        return UNDEFINED
    raw_data = group[name][()]
    size = raw_data.size
    if size is not None and size > 0:
        return raw_data.tobytes().rstrip(b"\x00")
    else:
        return None


def write_string(group: "Group", name: str, data: str, encoding="utf-8"):
    saved_data = np.bytes_([data]) if data is not None else []
    return group.create_dataset(
        name, data=saved_data, compression=COMPRESSION, dtype=h5py.string_dtype(encoding=encoding)
    )


def read_string(group: "Group", name: str):
    if not name in group:
        return UNDEFINED
    raw_data = group[name][()]
    size = raw_data.size
    if size is not None and size > 0:
        return np.bytes_(raw_data.flat[0]).decode()
    else:
        return None


def write_fitproblem(group: "Group", name: str, fitProblem: "bumps.fitproblem.FitProblem", serializer: SERIALIZERS):
    serialized = serialize_problem(fitProblem, serializer) if fitProblem is not None else None
    dset = write_bytes_data(group, name, serialized)
    return dset


def read_fitproblem(group: "Group", name: str, serializer: SERIALIZERS):
    if not name in group:
        return UNDEFINED
    serialized = read_bytes_data(group, name)
    fitProblem = deserialize_problem(serialized, serializer) if serialized is not None else None
    return fitProblem


def write_json(group: "Group", name: str, data):
    serialized = json.dumps(data) if data is not None else None
    dset = write_string(group, name, serialized.encode())
    return dset


def read_json(group: "Group", name: str):
    if not name in group:
        return UNDEFINED
    serialized = read_string(group, name)
    try:
        # if JSON fails to load, then just return None
        result = json.loads(serialized) if serialized is not None else None
    except Exception:
        result = None
    return result


def write_ndarray(group: "Group", name: str, data: Optional[np.ndarray], dtype=UNCERTAINTY_DTYPE):
    saved_data = data if data is not None else []
    return group.create_dataset(name, data=saved_data, dtype=dtype, compression=UNCERTAINTY_COMPRESSION)


def read_ndarray(group: "Group", name: str):
    if not name in group:
        return UNDEFINED
    raw_data = group[name][()]
    size = raw_data.size
    if size is not None and size > 0:
        return raw_data
    else:
        return None


class StringAttribute:
    @classmethod
    def serialize(value, obj=None):
        return json.dumps(value)

    @classmethod
    def deserialize(value, obj=None):
        return json.loads(value) if value else None


class ProblemState:
    fitProblem: Optional["bumps.fitproblem.FitProblem"] = None
    serializer: Optional[SERIALIZERS] = None

    def write(self, parent: "Group"):
        group = parent.require_group("problem")
        write_fitproblem(group, "fitProblem", self.fitProblem, self.serializer)
        write_string(group, "serializer", self.serializer)
        write_json(group, "libraries", get_libraries(self.fitProblem))
        # write_json(group, 'pathlist', self.pathlist)
        # write_string(group, 'filename', self.filename)

    def read(self, parent: "Group"):
        group = parent.require_group("problem")
        self.serializer = read_string(group, "serializer")
        self.fitProblem = read_fitproblem(group, "fitProblem", self.serializer)
        # self.pathlist = read_json(group, 'pathlist')
        # self.filename = read_string(group, 'filename')


class HistoryItem:
    problem: ProblemState
    fitting: Optional["FittingState"]
    timestamp: str
    label: str
    chisq_str: str
    keep: bool


class History:
    store: List[HistoryItem]

    def __init__(self):
        self.store = []

    def write(self, parent: "Group", include_uncertainty_state=True):
        group = parent.require_group("problem_history")
        for item in self.store:
            problem = item.problem
            fitting = item.fitting
            name = item.timestamp
            item_group = group.require_group(name)
            problem.write(item_group)
            fitting.write(item_group, include_uncertainty_state=include_uncertainty_state)
            item_group.attrs["chisq"] = item.chisq_str
            item_group.attrs["label"] = item.label
            item_group.attrs["keep"] = item.keep

    def read(self, parent: "Group"):
        group = parent.get("problem_history", [])
        self.store = []
        for name in group:
            item = HistoryItem()
            item_group = group[name]
            item.problem = ProblemState()
            item.fitting = FittingState()
            item.problem.read(item_group)
            item.fitting.read(item_group)
            item.label = item_group.attrs["label"]
            item.chisq_str = item_group.attrs["chisq"]
            # keep is a boolean, but h5py returns np.bool_ which is not JSON serializable
            item.keep = bool(item_group.attrs["keep"])
            item.timestamp = name
            self.store.append(item)

    def remove_item(self, timestamp: str):
        self.store = [item for item in self.store if item.timestamp != timestamp]

    def prune(self, target_length: int):
        # remove oldest items with keep=False until the length is target_length
        num_to_remove = len(self.store) - target_length
        if num_to_remove <= 0:
            return
        indices_to_remove = []
        for index, item in enumerate(self.store):
            if not item.keep:
                indices_to_remove.append(index)
                num_to_remove -= 1
                if num_to_remove == 0:
                    break
        for index in reversed(indices_to_remove):
            self.store.pop(index)

    def add_item(self, item: HistoryItem, target_length: int):
        self.prune(target_length)
        self.store.append(item)

    def list(self):
        return [
            dict(
                timestamp=item.timestamp,
                label=item.label,
                chisq_str=item.chisq_str,
                keep=item.keep,
                has_population=(item.fitting.population is not None),
                has_uncertainty=(item.fitting.uncertainty_state is not None),
            )
            for item in self.store
        ]

    def set_keep(self, timestamp: str, keep: bool):
        for item in self.store:
            if item.timestamp == timestamp:
                item.keep = keep
                return

    def update_label(self, timestamp: str, label: str):
        for item in self.store:
            if item.timestamp == timestamp:
                item.label = label
                return


class UncertaintyStateStorage:
    AR: Optional["np.ndarray"] = None
    gen_draws: Optional["np.ndarray"] = None
    labels: Optional["np.ndarray"] = None
    thin_draws: Optional["np.ndarray"] = None
    gen_logp: Optional["np.ndarray"] = None
    thin_logp: Optional["np.ndarray"] = None
    thin_point: Optional["np.ndarray"] = None
    update_CR_weight: Optional["np.ndarray"] = None
    update_draws: Optional["np.ndarray"] = None
    good_chains: Optional["np.ndarray"] = None

    def write(self, parent: "Group"):
        group = parent.require_group("uncertainty_state")
        for attrname in [
            "AR",
            "gen_draws",
            "thin_draws",
            "gen_logp",
            "thin_logp",
            "thin_point",
            "update_CR_weight",
            "update_draws",
            "good_chains",
        ]:
            write_ndarray(group, attrname, getattr(self, attrname), dtype=UNCERTAINTY_DTYPE)
        write_ndarray(group, "labels", self.labels, dtype=LABEL_DTYPE)

    def read(self, parent: "Group"):
        group = parent["uncertainty_state"]
        for attrname in [
            "AR",
            "gen_draws",
            "labels",
            "thin_draws",
            "gen_logp",
            "thin_logp",
            "thin_point",
            "update_CR_weight",
            "update_draws",
            "good_chains",
        ]:
            setattr(self, attrname, read_ndarray(group, attrname))


class FittingState:
    population: Optional[List] = None
    uncertainty_state: Optional["bumps.dream.state.MCMCDraw"] = None

    def write(self, parent: "Group", include_uncertainty_state=True):
        group = parent.require_group("fitting")
        write_ndarray(group, "population", self.population)
        uncertainty_state_storage = UncertaintyStateStorage()
        uncertainty_state = self.uncertainty_state
        if uncertainty_state is not None and include_uncertainty_state:
            write_uncertainty_state(uncertainty_state, uncertainty_state_storage)
            uncertainty_state_storage.write(group)

    def read(self, parent: "Group"):
        group = parent["fitting"]
        population = read_ndarray(group, "population")
        self.population = population
        if "uncertainty_state" in group:
            uncertainty_state_storage = UncertaintyStateStorage()
            uncertainty_state_storage.read(group)
            self.uncertainty_state = read_uncertainty_state(uncertainty_state_storage)


class State:
    # These attributes are ephemeral, not to be serialized/stored:
    hostname: str
    port: int
    parallel: int = 0
    fit_thread: Optional["FitThread"] = None
    fit_abort: Optional[Event] = None
    fit_abort_event: Event
    fit_complete_event: Event
    fit_uncertainty_final: Event
    fit_enabled: Event
    calling_loop: Optional[asyncio.AbstractEventLoop] = None
    base_path: str = ""

    # State to be stored:
    problem: ProblemState
    fitting: FittingState
    history: History
    topics: Dict["TopicNameType", "deque[Dict]"]
    shared: "SharedState"

    def __init__(self):
        self.problem = ProblemState()
        self.fitting = FittingState()
        self.history = History()
        self.fit_abort_event = Event()
        self.fit_complete_event = Event()
        self.fit_uncertainty_final = Event()
        self.topics = {
            "log": deque([]),
        }
        self.shared = SharedState()

    def __enter__(self):
        return self

    def setup_backing(self, session_file_name: str, session_pathlist: List[str], read_only: bool = False):
        if not read_only:
            self.shared.session_output_file = dict(filename=session_file_name, pathlist=session_pathlist)
        if session_file_name is not None:
            full_path = Path(*session_pathlist) / session_file_name
            if full_path.exists():
                self.read_session_file(full_path)
            else:
                self.save()

    def save_to_history(
        self, label: str, include_population: bool = False, include_uncertainty: bool = False, keep: bool = False
    ):
        if self.problem.fitProblem is None:
            return
        item = HistoryItem()
        item.problem = deepcopy(self.problem)
        item.fitting = FittingState()
        if include_population:
            item.fitting.population = self.fitting.population
        if include_uncertainty:
            # uncertainty_state is not mutated, so we can safely use it here
            # rather than making a memory-hogging copy on first use.
            item.fitting.uncertainty_state = self.fitting.uncertainty_state

        item.timestamp = str(datetime.now())
        item.label = label
        item.chisq_str = item.problem.fitProblem.chisq_str()
        item.keep = keep
        self.history.add_item(item, self.shared.autosave_history_length - 1)
        self.shared.updated_history = now_string()

    def get_history(self):
        return dict(problem_history=self.history.list())

    def remove_history_item(self, timestamp: str):
        self.history.remove_item(timestamp)

    def reload_history_item(self, timestamp: str):
        for item in self.history.store:
            if item.timestamp == timestamp:
                print("problem found!", timestamp)
                problem_state = item.problem
                print("chisq of found item: ", problem_state.fitProblem.chisq_str())
                self.problem = deepcopy(problem_state)
                self.shared.updated_model = now_string()
                self.shared.updated_parameters = now_string()
                if item.fitting.population is not None:
                    self.fitting.population = item.fitting.population
                    self.shared.updated_convergence = now_string()
                if item.fitting.uncertainty_state is not None:
                    self.fitting.uncertainty_state = item.fitting.uncertainty_state
                    self.shared.updated_uncertainty = now_string()
                return
        raise ValueError(f"Could not find history item with timestamp {timestamp}")

    def autosave(self):
        if self.shared.autosave_session:
            self.save()

    def save(self):
        if self.shared.session_output_file not in [None, UNDEFINED]:
            pathlist = self.shared.session_output_file["pathlist"]
            filename = self.shared.session_output_file["filename"]
            full_path = Path(*pathlist) / filename
            self.write_session_file(full_path)

    def write_session_file(self, session_fullpath: str):
        # Session filename is assumed to be a full path
        tmp_fd, tmp_name = tempfile.mkstemp(
            dir=Path(session_fullpath).parent, prefix=Path(session_fullpath).name, suffix=".tmp"
        )
        with os.fdopen(tmp_fd, "w+b") as output_file:
            with h5py.File(output_file, "w") as root_group:
                self.problem.write(root_group)
                self.fitting.write(root_group)
                self.history.write(root_group)
                self.write_topics(root_group)
                self.shared.write(root_group)
        shutil.move(tmp_name, session_fullpath)
        os.chmod(session_fullpath, 0o644)

    def read_session_file(self, session_fullpath: str, read_problem: bool = True, read_fitstate: bool = True):
        try:
            with h5py.File(session_fullpath, "r") as root_group:
                if read_problem:
                    self.problem.read(root_group)
                if read_fitstate:
                    self.fitting.read(root_group)
                self.history.read(root_group)
                self.read_topics(root_group)
                self.shared.read(root_group)
        except Exception as e:
            logger.warning(f"could not load session file {session_fullpath} because of {e}")

    def read_problem_from_session(self, session_fullpath: str):
        try:
            with h5py.File(session_fullpath, "r") as root_group:
                self.problem.read(root_group)
        except Exception as e:
            logger.warning(f"could not load fitProblem from {session_fullpath} because of {e}")

    def read_fitstate_from_session(self, session_fullpath: str):
        try:
            with h5py.File(session_fullpath, "r") as root_group:
                self.fitting.read(root_group)
        except Exception as e:
            logger.warning(f"could not load fit state from {session_fullpath} because of {e}")

    def write_topics(self, parent: "Group"):
        group = parent.require_group("topics")
        for topic, messages in self.topics.items():
            write_json(group, topic, list(messages))

    def read_topics(self, parent: "Group"):
        group = parent.require_group("topics")
        for topic in group:
            topic_data = read_json(group, topic)
            topic_data = np.array([topic_data]).flatten()
            if topic_data is not None and topic in self.topics:
                self.topics[topic].extend(topic_data)

    def get_last_message(self, topic: "TopicNameType"):
        return self.topics[topic][-1] if len(self.topics[topic]) > 0 else None

    def cleanup(self):
        pass

    def __del__(self):
        self.cleanup()

    async def async_cleanup(self):
        self.cleanup()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cleanup()


def write_uncertainty_state(state: "MCMCDraw", storage: UncertaintyStateStorage):
    # Build 2-D data structures
    storage.gen_draws, storage.gen_logp = state.logp(full=True)
    _, storage.AR = state.acceptance_rate()

    storage.thin_draws, storage.thin_point, storage.thin_logp = state.chains()
    storage.update_draws, storage.update_CR_weight = state.CR_weight()
    storage.labels = np.array(state.labels, dtype=LABEL_DTYPE)
    good_chains = state._good_chains
    storage.good_chains = None if isinstance(good_chains, slice) else good_chains


def read_uncertainty_state(loaded: UncertaintyStateStorage, skip=0, report=0, derived_vars=0):
    # Guess dimensions
    Ngen = loaded.gen_draws.shape[0]
    thinning = 1
    Nthin, Npop, Nvar = loaded.thin_point.shape
    Nupdate, Ncr = loaded.update_CR_weight.shape
    Nthin -= skip
    good_chains = loaded.good_chains

    # Create empty draw and fill it with loaded data
    state = MCMCDraw(0, 0, 0, 0, 0, 0, thinning)
    state.draws = Ngen * Npop
    state.labels = [label.decode() for label in loaded.labels]
    state.generation = Ngen
    state._gen_index = 0
    state._gen_draws = loaded.gen_draws
    state._gen_acceptance_rate = loaded.AR
    state._gen_logp = loaded.gen_logp
    state.thinning = thinning
    state._thin_count = Ngen // thinning
    state._thin_index = 0
    state._thin_draws = loaded.thin_draws
    state._thin_logp = loaded.thin_logp
    state._thin_point = loaded.thin_point
    state._gen_current = state._thin_point[-1].copy()
    state._update_count = Nupdate
    state._update_index = 0
    state._update_draws = loaded.update_draws
    state._update_CR_weight = loaded.update_CR_weight
    state._outliers = []

    bestidx = np.unravel_index(np.argmax(loaded.thin_logp), loaded.thin_logp.shape)
    state._best_logp = loaded.thin_logp[bestidx]
    state._best_x = loaded.thin_point[bestidx]
    state._best_gen = 0

    state._good_chains = (
        slice(None, None) if (good_chains is None or good_chains is UNDEFINED) else good_chains.astype(int)
    )

    return state


class ActiveFit(TypedDict):
    fitter_id: str
    options: Dict[str, Any]
    num_steps: int
    step: int
    chisq: str
    value: float


class FileInfo(TypedDict):
    filename: str
    pathlist: List[str]


Timestamp = NewType("Timestamp", str)


@dataclass
class SharedState:
    updated_convergence: Union[UNDEFINED_TYPE, Timestamp] = UNDEFINED
    updated_uncertainty: Union[UNDEFINED_TYPE, Timestamp] = UNDEFINED
    updated_parameters: Union[UNDEFINED_TYPE, Timestamp] = UNDEFINED
    updated_model: Union[UNDEFINED_TYPE, Timestamp] = UNDEFINED
    updated_history: Union[UNDEFINED_TYPE, Timestamp] = UNDEFINED
    selected_fitter: Union[UNDEFINED_TYPE, str] = UNDEFINED
    fitter_settings: Union[UNDEFINED_TYPE, Dict[str, Dict]] = UNDEFINED
    active_fit: Union[UNDEFINED_TYPE, ActiveFit] = UNDEFINED
    model_file: Union[UNDEFINED_TYPE, FileInfo] = UNDEFINED
    model_loaded: Union[UNDEFINED_TYPE, bool] = UNDEFINED
    session_output_file: Union[UNDEFINED_TYPE, FileInfo] = UNDEFINED
    autosave_session: bool = False
    autosave_session_interval: int = 300
    autosave_history: bool = True
    autosave_history_length: int = 10

    _not_reloaded = ["active_fit", "autosave_session", "session_output_file", "_notification_callbacks"]
    _notification_callbacks: Dict[str, Callable[[str, Any], Awaitable[None]]] = field(default_factory=dict)

    def __setattr__(self, name: str, value):
        super().__setattr__(name, value)
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # no event loop running, so no need to notify
            return
        if loop.is_running() and hasattr(self, "_notification_callbacks"):
            for callback in self._notification_callbacks.values():
                loop.create_task(callback(name, value))

    async def set(self, name, value):
        super().__setattr__(name, value)
        for callback in self._notification_callbacks.values():
            await callback(name, value)

    async def get(self, name):
        return getattr(self, name, UNDEFINED)

    def write(self, parent: "Group"):
        group = parent.require_group("shared")
        for field in fields(self):
            if not field.name in self._not_reloaded:
                value = getattr(self, field.name)
                if value is not UNDEFINED:
                    write_json(group, field.name, value)

    def read(self, parent: "Group"):
        group = parent.get("shared")
        if group is None:
            return
        for field in fields(self):
            if not field.name in self._not_reloaded:
                setattr(self, field.name, read_json(group, field.name))



"""
Kolmogorov-Smirnov test for MCMC convergence.

Use the K-S tests to compare the distribution of values at the front of
the chain to that at the end of the chain.  If the distributions are
significantly different, then the MCMC chain has not converged.
"""

__all__ = ["ksmirnov"]

from numpy import reshape, apply_along_axis
from scipy.stats import ks_2samp


def ksmirnov(seq, portion=0.25, filter_order=15):
    """
    Kolmogorov-Smirnov test of similarity between the empirical distribution
    at the start and at the end of the chain.  Apply a median
    filter (filter=15) on neighbouring K-S values to reduce variation in
    the test statistic value.
    """
    chlen, nchains, nvars = seq.shape
    count = portion * chlen * nchains
    n = filter_order
    ks, p = apply_along_axis(lambda chain: _ksm(chain, n, count), 0, reshape(seq, (chlen * nchains, nvars)))
    return ks, p


def _ksm(chain, n, count):
    # return ks_2samp(chain[:count], chain[-count:])
    ks, p = zip(*[ks_2samp(chain[i : count + i], chain[-count - n + i : -n + i]) for i in range(n)])
    return sorted(ks)[(n - 1) // 2], sorted(p)[(n - 1) // 2]



"""
DiffeRential Evolution Adaptive Metropolis algorithm

DREAM runs multiple different chains simultaneously for global exploration,
and automatically tunes the scale and orientation of the proposal
distribution using differential evolution.  The algorithm maintains
detailed balance and ergodicity and works well and efficient for a large
range of problems, especially in the presence of high-dimensionality and
multimodality.

DREAM developed by Jasper A. Vrugt and Cajo ter Braak

This algorithm has been described in:

   Vrugt, J.A., C.J.F. ter Braak, M.P. Clark, J.M. Hyman, and B.A. Robinson,
      *Treatment of input uncertainty in hydrologic modeling: Doing hydrology
      backward with Markov chain Monte Carlo simulation*,
      Water Resources Research, 44, W00B09, 2008.
      `doi:10.1029/2007WR006720 <http://dx.doi.org/10.1029/2007WR006720>`_

   Vrugt, J.A., C.J.F. ter Braak, C.G.H. Diks, D. Higdon, B.A. Robinson,
       and J.M. Hyman,
       *Accelerating Markov chain Monte Carlo simulation by differential
       evolution with self-adaptive randomized subspace sampling*,
       International Journal of Nonlinear Sciences and Numerical Simulation,
       10(3), 271-288, 2009.

   Vrugt, J.A., C.J.F. ter Braak, H.V. Gupta, and B.A. Robinson,
       *Equifinality of formal (DREAM) and informal (GLUE) Bayesian approaches
       in hydrologic modeling*,
       Stochastic Environmental Research and Risk Assessment,
       1-16, 2009, In Press.
       `doi:10.1007/s00477-008-0274-y
       <http://dx.doi.org/10.1007/s00477-008-0274-y>`_

For more information please read:

   Ter Braak, C.J.F.,
       *A Markov Chain Monte Carlo version of the genetic algorithm Differential
       Evolution: easy Bayesian computing for real parameter spaces*,
       Stat. Comput., 16, 239 - 249, 2006.
       `doi:10.1007/s11222-006-8769-1
       <http://dx.doi.org/10.1007/s11222-006-8769-1>`_

   Vrugt, J.A., H.V. Gupta, W. Bouten and S. Sorooshian,
       *A Shuffled Complex Evolution Metropolis algorithm for optimization
       and uncertainty assessment of hydrologic model parameters*,
       Water Resour. Res., 39 (8), 1201, 2003.
       `doi:10.1029/2002WR001642 <http://dx.doi.org/10.1029/2002WR001642>`_

   Ter Braak, C.J.F., and J.A. Vrugt,
       *Differential Evolution Markov Chain with snooker updater
       and fewer chains*,
       Statistics and Computing, 2008.
       `doi:10.1007/s11222-008-9104-9
       <http://dx.doi.org/2008.10.1007/s11222-008-9104-9>`_

   Vrugt, J.A., C.J.F. ter Braak, and J.M. Hyman,
       *Differential evolution adaptive Metropolis with snooker update and
       sampling from past states*,
       SIAM journal on Optimization, 2009.

   Vrugt, J.A., C.J.F. ter Braak, and J.M. Hyman,
       *Parallel Markov chain Monte Carlo simulation on distributed computing
       networks using multi-try Metropolis with sampling from past states*,
       SIAM journal on Scientific Computing, 2009.

   G. Schoups, and J.A. Vrugt,
       *A formal likelihood function for Bayesian inference of hydrologic
       models with correlated, heteroscedastic and non-Gaussian errors*,
       Water Resources Research, 2010, In Press.

   G. Schoups, J.A. Vrugt, F. Fenicia, and N.C. van de Giesen,
       *Inaccurate numerical solution of hydrologic models corrupts efficiency
       and robustness of MCMC simulation*,
       Water Resources Research, 2010, In Press.

Copyright (c) 2008, Los Alamos National Security, LLC
All rights reserved.

Copyright 2008. Los Alamos National Security, LLC. This software was produced
under U.S. Government contract DE-AC52-06NA25396 for Los Alamos National
Laboratory (LANL), which is operated by Los Alamos National Security, LLC
for the U.S. Department of Energy. The U.S. Government has rights to use,
reproduce, and distribute this software.

NEITHER THE GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY, LLC MAKES ANY
WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF
THIS SOFTWARE.  If software is modified to produce derivative works, such
modified software should be clearly marked, so as not to confuse it with
the version available from LANL.

Additionally, redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of Los Alamos National Security, LLC, Los Alamos National
  Laboratory, LANL the U.S. Government, nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY LOS ALAMOS NATIONAL SECURITY, LLC AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL LOS ALAMOS NATIONAL
SECURITY, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

MATLAB code written by Jasper A. Vrugt, Center for NonLinear Studies (CNLS)

Written by Jasper A. Vrugt: vrugt@lanl.gov

Version 0.5: June 2008
Version 1.0: October 2008  Adaption updated and generalized CR implementation



2010-04-20 Paul Kienzle
* Convert to python
"""

__all__ = ["Dream"]

import sys
import time
from ctypes import c_double

import numpy as np

from .state import MCMCDraw
from .metropolis import metropolis, metropolis_dr, dr_step
from .crossover import AdaptiveCrossover, LogAdaptiveCrossover
from .diffev import de_step
from .bounds import make_bounds_handler
from .compiled import dll
from .util import rng
from .convergence import ks_converged

# Everything should be available in state, but lets be lazy for now
LAST_TIME = 0


def console_monitor(state, pop, logp):
    """
    Print progress of fit on the console.
    """
    global LAST_TIME
    if state.generation == 1:
        print("#gen", "logp(x)", " ".join("<%s>" % par for par in state.labels))
        LAST_TIME = time.time()

    current_time = time.time()
    if current_time >= LAST_TIME + 1:
        LAST_TIME = current_time
        print(state.generation, state._best_logp, " ".join("%.15g" % v for v in state._best_x))
        sys.stdout.flush()


class Dream(object):
    """
    Data structure containing the details of the running DREAM analysis code.
    """

    model = None
    # Sampling parameters
    burn = 0
    draws = 100000
    thinning = 1
    # TODO: change the default outlier test to IQR and control with options
    outlier_test = "none"
    population = None
    #: convergence criteria
    alpha = 0.01
    # DE parameters
    DE_steps = 10
    DE_pairs = 3
    DE_eps = 0.05
    DE_snooker_rate = 0.1
    DE_noise = 1e-6
    bounds_style = "reflect"
    # Crossover parameters
    CR = None
    CR_spacing = "linear"  # 'log' or 'linear'
    # Delay rejection parameters
    use_delayed_rejection = False
    DR_scale = 1  # 1-sigma step size using cov of population
    # Local optimizer best fit injection  The optimizer has
    # the following interface:
    #    x, fx = goalseek(mapper, bounds_handler, pop, fpop)
    # where:
    #    x, fx are the local optimum point and its value
    #    pop is the starting population
    #    fpop is the nllf for each point in pop
    #    mapper is a function which takes pop and returns fpop
    #    bounds_handler takes pop and forces all points into the range
    goalseek_optimizer = None
    goalseek_interval = 1e100  # close enough to never
    goalseek_minburn = 1000
    state = None  # type: MCMCDraw

    def __init__(self, **kw):
        self.monitor = console_monitor
        for k, v in kw.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise TypeError("Unknown attribute " + k)

        self._initialized = False

    def sample(self, state=None, abort_test=lambda: False):
        """
        Pull the requisite number of samples from the distribution
        """
        if not self._initialized:
            self._initialized = True
        self.state = state
        try:
            _run_dream(self, abort_test=abort_test)
        except KeyboardInterrupt:
            pass
        return self.state


def _run_dream(dream, abort_test=lambda: False):
    """
    Collect posterior distribution samples using DREAM sampler.
    """

    if dll:
        dll.rand_init(rng.randint(1e9))
    # Step 1: Sample s points in the parameter space
    # [PAK] I moved this out of dream so that the user can use whatever
    # complicated sampling scheme they want.  Unfortunately, this means
    # the user needs to know some complex sampling scheme.
    if dream.population is None:
        raise ValueError("initial population not defined")

    # Remember the problem dimensions
    n_gen, n_chain, n_var = dream.population.shape
    n_pop = n_gen * n_chain

    if dream.CR is None:
        if dream.CR_spacing == "log":
            dream.CR = LogAdaptiveCrossover(n_var)
        else:  # linear
            dream.CR = AdaptiveCrossover(3)

    # Step 2: Calculate posterior density associated with each value in x
    apply_bounds = make_bounds_handler(dream.model.bounds, style=dream.bounds_style)

    # Record initial state
    allocate_state(dream)
    state = dream.state
    state.labels = dream.model.labels
    previous_draws = state.draws
    if previous_draws:
        x, logp = state._last_gen()
    else:
        # No initial state, so evaluate initial population
        for x in dream.population:
            apply_bounds(x)
            # ********************** MAP *****************************
            logp = dream.model.map(x)
            state._generation(new_draws=n_chain, x=x, logp=logp, accept=n_chain, force_keep=True)
            dream.monitor(state, x, logp)

    state._update(CR_weight=dream.CR.weight)

    # Now start drawing samples
    # print "previous draws", previous_draws, "new draws", dream.draws+dream.burn
    last_goalseek = (dream.draws + dream.burn) / n_pop - dream.goalseek_minburn
    next_goalseek = state.generation + dream.goalseek_interval if dream.goalseek_optimizer else 1e100

    if dll:
        xtry = np.empty((n_chain, n_var), "d")
        step_alpha = np.empty(n_chain, "d")
        CR_used = np.empty(n_chain, "d")
    scale = 1.0
    # serial_time = parallel_time = 0.
    # last_time = time.time()

    # Make sure that the pop we are drawing doesn't need to be copied before
    # sending it to the compiled C code.
    pop = state._draw_pop()
    assert pop.ctypes.data == np.ascontiguousarray(pop).ctypes.data

    frame = 0
    next_outlier_test = max(state.Ngen, 2 * state.Ngen - 10)
    next_convergence_test = state.Ngen
    final_gen = dream.draws + dream.burn
    while state.draws < final_gen:
        # Age the population using differential evolution
        dream.CR.reset()
        CR = np.ascontiguousarray(np.vstack((dream.CR.CR, dream.CR.weight)).T, "d")
        for gen in range(dream.DE_steps):
            # Define the current locations and associated posterior densities
            xold, logp_old = x, logp
            pop = state._draw_pop()
            # print(pop.ctypes.data, np.ascontiguousarray(pop).ctypes.data)
            # print(pop)
            # print("gen", gen, pop.shape)

            # Generate candidates for each sequence
            if dll is None:
                xtry, step_alpha, CR_used = de_step(
                    n_chain,
                    pop,
                    CR,
                    max_pairs=dream.DE_pairs,
                    eps=dream.DE_eps,
                    snooker_rate=dream.DE_snooker_rate,
                    noise=dream.DE_noise,
                    scale=scale,
                )
            else:
                dll.de_step(
                    n_chain,
                    n_var,
                    len(CR),
                    pop.ctypes,
                    CR.ctypes,
                    dream.DE_pairs,
                    c_double(dream.DE_eps),
                    c_double(dream.DE_snooker_rate),
                    c_double(dream.DE_noise),
                    c_double(scale),
                    xtry.ctypes,
                    step_alpha.ctypes,
                    CR_used.ctypes,
                )
            # print("try", xtry)

            # PAK: Try a local optimizer every N generations
            if next_goalseek <= state.generation <= last_goalseek:
                best, logp_best = dream.goalseek_optimizer(dream.model.map, apply_bounds, xold, logp_old)
                xtry[0] = best
                # Note: it is slightly inefficient to throw away logp_best,
                # but it makes the the code cleaner if we do
                next_goalseek += dream.goalseek_interval

            # Compute the likelihood of the candidates
            apply_bounds(xtry)
            # ********************** MAP *****************************
            # next_time = time.time()
            # serial_time += next_time - last_time
            # last_time = next_time
            logp_try = dream.model.map(xtry)
            # next_time = time.time()
            # parallel_time  += next_time - last_time
            # last_time = next_time
            draws = len(logp_try)

            # Apply the metropolis acceptance/rejection rule
            x, logp, alpha, accept = metropolis(xtry, logp_try, xold, logp_old, step_alpha)

            # Process delayed rejection
            # PAK NOTE: this updates according to the covariance matrix of the
            # current sample, which may be useful on unimodal systems, but
            # doesn't seem to be of any value in general; the DREAM papers
            # found that the acceptance rate did indeed improve with delayed
            # rejection, but the overall performance did not.  Worse, this
            # requires a linear system solution O(nPop^3) which can be near
            # singular for complex posterior distributions.
            if dream.use_delayed_rejection and not accept.all():
                # Generate alternate candidates using the covariance of xold
                xdr, r = dr_step(x=xold, scale=dream.DR_scale)

                # Compute the likelihood of the new candidates
                reject = ~accept
                apply_bounds(xdr)
                # ********************** MAP *****************************
                logp_xdr = dream.model.map(xdr[reject])
                draws += len(logp_xdr)

                # Apply the metropolis delayed rejection rule.
                x[reject], logp[reject], alpha[reject], accept[reject] = metropolis_dr(
                    xtry[reject],
                    logp_try[reject],
                    x[reject],
                    logp[reject],
                    xold[reject],
                    logp_old[reject],
                    alpha[reject],
                    r,
                )

            # els = zip(logp_old, logp_try, logp, x[:, -2], x[:, -1], accept)
            # print "pop", "\n ".join((("%12.3e "*(len(el)-1))%el[:-1])
            #                       +("T " if el[-3]<=el[-2] else "  ")
            #                       +("accept" if el[-1] else "")
            #                       for el in els)

            # Update Sequences with the new population.
            state._generation(draws, x, logp, accept)
            # ********************** NOTIFY **************************
            dream.monitor(state, x, logp)
            # print state.generation, ":", state._best_logp

            # Keep track of which CR ratios were successful
            if state.draws <= dream.burn:
                dream.CR.update(xold, x, CR_used)

            if abort_test():
                break

        # print("serial&parallel",serial_time,parallel_time)
        # End of differential evolution aging
        # ---------------------------------------------------------------------

        # Save update information
        state._update(CR_weight=dream.CR.weight)

        if abort_test():
            break

        # if state.draws <= 0.1 * dream.draws:
        if state.draws <= dream.burn:
            # Adapt the crossover ratio, but only during burn-in.
            dream.CR.adapt()

        if False:
            # Suppress scale update until we have a chance to verify that it
            # doesn't skew the resulting statistics.
            _, r = state.acceptance_rate()
            ravg = np.mean(r[-dream.DE_steps :])
            if ravg > 0.4:
                scale *= 1.01
            elif ravg < 0.2:
                scale /= 1.01

        if state.generation >= next_convergence_test:
            converged = ks_converged(state, alpha=dream.alpha)
        else:
            converged = False

        # See whether there are any outlier chains, and remove them
        # Only do this once per frame, and only if there is some time
        # left to adapt the distribution.  Also do it if we believe
        # that we have converged.
        if (converged or state.generation >= next_outlier_test) and state.generation + state.Ngen < final_gen:
            outliers = state.remove_outliers(x, logp, dream.outlier_test)
            next_outlier_test = state.generation + 2 * state.Ngen
            if len(outliers):
                # TODO: use monitors to report arbitrary information
                print("step %d trimmed %d outliers from %d chains" % (state.generation, len(outliers), len(logp)))
                next_convergence_test = state.generation + state.Ngen
                converged = False

        if converged:
            # _show_logp_frame(dream, state, frame+1)
            break

        # Draw the next frame (for debugging...)
        next_frame = state.generation // state.Ngen
        # if frame != next_frame: _show_logp_frame(dream, state, next_frame)
        frame = next_frame


def _show_logp_frame(dream, state, frame):
    from pylab import clf, savefig
    from . import views

    clf()
    views.plot_logp(state)
    savefig("logp%03d.png" % frame)
    clf()


def allocate_state(dream):
    """
    Estimate the size of the output
    """
    # Determine problem dimensions from the initial population
    n_pop, n_chain, n_var = dream.population.shape
    steps = dream.DE_steps
    thinning = dream.thinning
    n_cr = len(dream.CR.CR)
    draws = dream.draws

    n_update = int(draws / (steps * n_chain)) + 1
    n_gen = n_update * steps
    n_thin = int(n_gen / thinning) + 1

    # print("new state", n_var, n_chain, n_cr, n_gen, n_thin, n_update, draws, steps)
    if dream.state is not None:
        # print("existing", dream.state.Nvar, dream.state.Npop, dream.state.Ncr)
        dream.state.resize(n_gen, n_thin, n_update, n_var, n_chain, n_cr, thinning)
    else:
        dream.state = MCMCDraw(n_gen, n_thin, n_update, n_var, n_chain, n_cr, thinning)



"""
Ccorana's function

References::
    [1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
    Heuristic for Global Optimization over Continuous Spaces. Journal of Global
    Optimization 11: 341-359, 1997.

    [2] Storn, R. and Price, K.
    (Same title as above, but as a technical report.)
    http://www.icsi.berkeley.edu/~storn/deshort1.ps
"""

from math import pow

from numpy import sign, floor, inf


def corana(coeffs):
    """
    evaluates the Corana function for a list of coeffs

    minimum is f(x)=0.0 at xi=0.0
    """
    d = [1.0, 1000.0, 10.0, 100.0]
    x = coeffs
    r = 0
    for xj, dj in zip(x, d):
        zj = floor(abs(xj / 0.2) + 0.49999) * sign(xj) * 0.2
        if abs(xj - zj) < 0.05:
            r += 0.15 * pow(zj - 0.05 * sign(zj), 2) * dj
        else:
            r += dj * xj**2
        return r


def corana1d(x):
    """Corana in 1D; coeffs = (x,0,0,0)"""
    return corana([x[0], 0, 0, 0])


def corana2d(x):
    """Corana in 2D; coeffs = (x,0,y,0)"""
    return corana([x[0], 0, x[1], 0])


def corana3d(x):
    """Corana in 3D; coeffs = (x,0,y,z)"""
    return corana([x[0], 0, x[1], x[2]])


Po = [1, 1, 1, 1]
Plo = [-inf, -inf, -inf, -inf]
Phi = [inf, inf, inf, inf]


def s(V, *args):
    retval = []
    for i in args:
        retval.append(V[i])
    return retval


from .model import Function

corana1d = Function(f=corana1d, limits=(s(Plo, 0), s(Phi, 0)), start=s(Po, 0))
corana2d = Function(f=corana2d, limits=(s(Plo, 0, 2), s(Phi, 0, 2)), start=s(Po, 0, 2))
corana3d = Function(f=corana3d, limits=(s(Plo, 0, 2, 3), s(Phi, 0, 2, 3)), start=s(Po, 0, 2, 3))
corana4d = Function(f=corana, limits=(Plo, Phi), start=Po)
# End of file



"""
Data handling utilities.
"""

import os
import gzip
from contextlib import contextmanager

import numpy as np
from numpy import inf, nan

__all__ = ["indfloat", "parse_file"]


def parse_multi(file, keysep=None, sep=None, comment="#"):
    """
    Parse a multi-part file.

    Return a list of (header, data) pairs, where header is a key: value
    dictionary and data is a numpy array.

    The header section is list of key value pairs, with the *comment* character
    at the start of each line.  Key and value will be separated by *keysep*,
    or by spaces if *keysep = None*.  The data section is a sequence of
    floating point numbers separated by *sep*, or by spaces if *sep* is None.
    inf and nan are parsed as inf and nan.  Comments at the end of the data
    line will be ignored.  Data points can be commented out by including
    a comment character at the start of the data line, assuming the next
    character is a digit, plus, or decimal separator.

    Quotes around keys are removed, but not around values.  Use
    :func:`strip_quotes` to remove them if they are present.  This is different
    from the :func:`parse_file` interface, which strips quotes around values.
    The new interface allows *json.loads()* calls on values if values are
    stored as *key: json.dumps(value)*.

    Special hack for binned data: if the first column contains bin edges, then
    the last row will only have the bin edge.  To make the array square,
    we replace the bin edges with bin centers.  The original bins can be
    found in the header using the 'bins' key (unless that key already exists
    in the header, in which case the key will be ignored).
    """
    parts = []
    with maybe_open(file) as fh:
        while True:
            header, data, bins = _read_part(fh, comment=comment, multi_part=True, col_sep=sep, key_sep=keysep)
            if header is None:
                break
            if bins is not None:
                header.setdefault("bins", bins)
            parts.append((header, data))
    return parts


def parse_file(file, keysep=None, sep=None, comment="#"):
    """
    Parse a file into a header and data.

    Return a (header, data) pair, where header is a key: value
    dictionary and data is a numpy array.

    The header section is list of key value pairs, with the *comment* character
    at the start of each line.  Key and value will be separated by *keysep*,
    or by spaces if *keysep = None*.  The data section is a sequence of
    floating point numbers separated by *sep*, or by spaces if *sep* is None.
    inf and nan are parsed as inf and nan.  Comments at the end of the data
    line will be ignored.  Data points can be commented out by including
    a comment character at the start of the data line, assuming the next
    character is a digit, plus, or decimal separator.

    Quotes around keys are removed.  For compatibility with the old interface,
    quotes around values are removed as well.

    Special hack for binned data: if the first column contains bin edges, then
    the last row will only have the bin edge.  To make the array square,
    we replace the bin edges with bin centers.  The original bins can be
    found in the header using the 'bins' key (unless that key already exists
    in the header, in which case the key will be ignored).
    """
    with maybe_open(file) as fh:
        header, data, bins = _read_part(fh, comment=comment, multi_part=False, col_sep=sep, key_sep=keysep)
    if header is None:
        raise IOError("data file is empty")
    # compatibility: strip quotes from values in key-value pairs
    header = dict((k, strip_quotes(v)) for k, v in header.items())
    if bins is not None:
        header.setdefault("bins", bins)
    return header, data


def _read_part(fh, key_sep=None, col_sep=None, comment="#", multi_part=False):
    header = {}
    data = []
    iseof = True
    for line in fh:
        # Blank lines indicate a section break.
        if not line.strip():
            # Skip blank lines if we are parsing the data as a single part file
            # or if the "section" has no data due to blank lines in the header.
            if not multi_part or not data:
                continue
            # If we are at the beginning of a section, then iseof is True and
            # continuing to the next loop iteration will skip them. If we have
            # already consumed some non-blank lines, then iseof will be false,
            # and we need to break this section of the data.  If we have blank
            # lines at the end of the file, we will never set iseof to False
            # and they will be ignored.
            if iseof:
                continue
            break

        # Line is not blank, so process it.
        columns, key, value = _parse_line(line, comment=comment, col_sep=col_sep, key_sep=key_sep)
        if columns:
            data.append([indfloat(v) for v in columns])
        if key is not None:
            if key in header:
                header[key] = "\n".join((header[key], value))
            else:
                header[key] = value

        # We have processed some data, so
        iseof = False

    if iseof:
        return None, None, None

    # print data
    # print "\n".join(k+":"+v for k,v in header.items())
    if len(data) and len(data[-1]) == 1:
        # For TOF data, the first column is the bin edge, which has one
        # more row than the remaining columns; fill those columns with
        # bin centers instead
        last_edge = data[-1][0]
        data = np.array(data[:-1]).T
        edges = np.hstack((data[0], last_edge))
        data[0] = 0.5 * (edges[:-1] + edges[1:])
        bins = edges
    else:
        data = np.array(data).T
        bins = None

    return header, data, bins


@contextmanager
def maybe_open(file_or_path):
    """
    A context manager for file opening, given as a file path or an open handle.

    If *file_or_path* is a string ending in ".gz" then open with gzip.
    """
    if hasattr(file_or_path, "readline"):
        # If it is a file handle, yield it and return without closing.
        fh = file_or_path
        yield fh
    else:
        # Otherwise it should be a path. Make sure it is at least a string.
        if not string_like(file_or_path):
            raise ValueError("file must be a name or a file handle")
        # Open file; if name ends in .gz then assume it is compressed.
        path = os.path.expanduser(file_or_path)
        fh = gzip.open(path) if path.endswith(".gz") else open(path)
        try:
            yield fh
        finally:
            fh.close()


def string_like(s):
    """
    Return True if s operates like a string.
    """
    try:
        s + ""
    except Exception:
        return False
    return True


def _parse_line(line, key_sep=None, col_sep=None, comment="#"):
    # Find location of the comment character on the line
    idx = line.find(comment)

    # If the line does not contain a comment character or if the comment
    # character is not in the first column, then this is a data line which
    # should be returned as a sequence of text columns separated by spaces.
    # The caller can turn the columns into numbers or leave them as strings.
    # Data on the line after the comment character is ignored.
    # TODO: allow quoted strings or backslash escaped spaces for text columns
    if idx != 0:
        if idx > 0:
            return line[:idx].split(col_sep), None, ""
        else:
            return line.split(col_sep), None, ""

    # Split line on key separator
    parts = [p.strip() for p in line[1:].split(key_sep, 1)]
    key, value = parts if len(parts) > 1 else (parts[0], "")
    key = strip_quotes(key)

    # If key is a number assume it is simply a commented out data point
    if len(key) and (key[0] in ".-+0123456789" or key == "inf" or key == "nan"):
        return [], None, None

    return [], key, value


def strip_quotes(s):
    return s[1:-1] if len(s) and s[0] in "'\"" and s[0] == s[-1] else s


INF_VALUES = set(("inf", "1/0", "1.#inf", "infinity"))
NAN_VALUES = set(("nan", "0/0", "1.#qnan", "na", "n/a"))


def indfloat(s):
    """
    Convert string to float, with support for inf and nan.

    Example::

        >>> from numpy import isinf, isnan
        >>> print(isinf(indfloat('inf')))
        True
        >>> print(isinf(indfloat('-inf')))
        True
        >>> print(isnan(indfloat('nan')))
        True
    """
    try:
        return float(s)
    except Exception:
        s = s.lower()
        if s in INF_VALUES:
            return inf
        elif s and s[0] == "-" and s[1:] in INF_VALUES:
            return -inf
        elif s in NAN_VALUES:
            return nan
        raise



# This program is public domain
# Author: Paul Kienzle, April 2010
"""
Mahalanobis distance calculator

Compute the
`Mahalanobis distance <https://en.wikipedia.org/wiki/Mahalanobis_distance>`_
between observations and a reference set.  The principle components of the
reference set define the basis of the space for the observations.  The simple
Euclidean distance is used within this space.
"""

__all__ = ["mahalanobis"]

from numpy import dot, mean, sum
from numpy.linalg import svd


def mahalanobis(Y, X):
    """
    Returns the distances of the observations from a reference set.

    Observations are stored in rows *Y* and the reference set in *X*.
    """

    M = mean(X, axis=0)  # mean
    Xc = X - mean(X, axis=0)  # center the reference
    W = dot(Xc.T, Xc) / (Xc.shape[0] - 1)  # covariance of reference
    Yc = Y - M  # center the observations
    # Distance is diag(Yc * inv(W) * Yc.H)
    # solve Wb = Yc.H using singular value decomposition because it is
    # the most accurate with numpy; QR decomposition does not use pivoting,
    # and is less accurate.  The built-in solve routine is between the two.
    u, s, vh = svd(W, 0)
    SVinv = vh.T.conj() / s
    Uy = dot(u.T.conj(), Yc.T.conj())
    b = dot(SVinv, Uy)

    D = sum(Yc.T * b, axis=0)  # compute distance
    return D


def test():
    from numpy import array
    from numpy.linalg import norm

    d = mahalanobis(array([[2, 3, 4], [2, 3, 4]]), array([[1, 0, 0], [2, 1, 0], [1, 1, 0], [2, 0, 1]]))
    assert norm(d - [290.25, 290.25]) < 1e-12, "diff=%s" % str(d - [290.25, 290.25])


if __name__ == "__main__":
    test()



"""
Differential evolution MCMC stepper.
"""

__all__ = ["de_step"]

from numpy import zeros, ones, empty, sqrt, sum
from numpy import where, array
from numpy.linalg import norm

from .util import draw, rng

try:
    # raise ImportError("skip numpy")
    from numba import njit, prange

    @njit(cache=True)
    def pchoice(choices, size=0, replace=True, p=None):
        if p is None:
            return rng.choice(choices, size=size, replace=replace)
        # TODO: if choices is an array, then shape should be an array
        result = empty(size, dtype=choices.dtype)
        num_choices = choices.shape[0]
        for index in prange(size):
            u = rng.rand()
            cdf = 0.0
            for k in range(num_choices):
                cdf += p[k]
                if u <= cdf:
                    result[index] = choices[k]
            # else: should never get here if choices sum to 1
        return result

except ImportError:

    def njit(*args, **kw):
        return lambda f: f

    prange = range
    pchoice = rng.choice

EPS = 1e-6
_DE, _SNOOKER, _DIRECT = 0, 1, 2


@njit(cache=True)
def de_step(Nchain, pop, CR, max_pairs=2, eps=0.05, snooker_rate=0.1, noise=1e-6, scale=1.0):
    """
    Generates offspring using METROPOLIS HASTINGS monte-carlo markov chain

    *Nchain* is the number of simultaneous changes that are running.

    *pop* is an array of shape [Npop x Nvar] providing the active points used
    to generate the next proposal for each chain. This may be larger than the
    Nchains if the caller is using ancestor generations for active population.
    The current population is assumed to be the first *Nchain* rows of *pip*..

    *CR* is an array of [Ncrossover x 2] crossover ratios with weights. The
    crossover ratio is the probability of selecting a particular dimension when
    generating the difference vector. The weights are used to select the
    crossover ratio. The weights are adjusted dynamically during the fit based
    on the acceptance rate of points generated with each crossover ratio.

    *max_pairs* determines the maximum number of pairs which contribute to the
    differential evolution step. The number of pairs is chosen at random, with
    the difference vectors between the pairs averaged when creating the DE step.

    *eps* determines the jitter added to the DE step.

    *snooker_rate* determines the probability of using the snooker stepper.
    Otherwise use DE stepper 80% of the time, or apply the difference between
    pairs the other 20% of the time.

    *scale=1* scales the difference vector (constant, not stochastic)

    *noise=1e-6* adds random noise to the non-zero components of the
    difference vector. This noise is relative rather than absolute to allow
    for parameter values far from 1.0. Noise is also scaled by *scale*.
    """
    Npop, Nvar = pop.shape

    # Initialize the delta update to zero
    delta_x = zeros((Nchain, Nvar))
    step_alpha = ones(Nchain)

    # Choose snooker, de or direct according` to snooker_rate, and 80:20
    # ratio of de to direct.
    u = rng.rand(Nchain)
    de_rate = 0.8 * (1 - snooker_rate)
    alg = zeros(Nchain, dtype="int")  # _DE = 0
    alg[u >= de_rate] = _SNOOKER
    alg[u >= de_rate + snooker_rate] = _DIRECT
    # alg = select([u < snooker_rate, u < snooker_rate+de_rate],
    #             [_SNOOKER, _DE], default=_DIRECT)
    # [PAK] CR selection moved from crossover into DE step
    CR_used = pchoice(CR[:, 0], size=Nchain, replace=True, p=CR[:, 1])

    # Chains evolve using information from other chains to create offspring
    for qq in prange(Nchain):
        if alg[qq] == _DE:  # Use DE with cross-over ratio
            # Select to number of vector pair differences to use in update
            # using k ~ discrete U[1, max pairs]
            k = rng.randint(max_pairs) + 1
            # [PAK: same as k = DEversion[qq, 1] in matlab version]

            # TODO: make sure we don't fail if too few chains.
            # Select 2*k members at random different from the current member
            perm = draw(2 * k, Npop - 1)
            # TODO: rewrite draw so that it accepts a not_matching int
            perm[perm >= qq] += 1
            r1, r2 = perm[:k], perm[k : 2 * k]

            # Select the dims to update based on the crossover ratio, making
            # sure at least one dim is selected
            vars = where(rng.rand(Nvar) <= CR_used[qq])[0]
            if len(vars) == 0:
                vars = array([rng.randint(Nvar)])
            # print("for chain", qq, CR_used[qq], "% update", vars)

            # Weight the size of the jump inversely proportional to the
            # number of contributions, both from the parameters being
            # updated and from the population defining the step direction.
            gamma = 2.38 / sqrt(2 * len(vars) * k)
            # [PAK: same as F=Table_JumpRate[len(vars), k] in matlab version]

            # Find and average step from the selected pairs
            step = sum(pop[r1] - pop[r2], axis=0)

            # Apply that step with F scaling and noise
            jiggle = 1 + eps * (2 * rng.rand(*step.shape) - 1)
            delta_x[qq, vars] = (jiggle * gamma * step)[vars]

        elif alg[qq] == _SNOOKER:  # Use snooker update
            # Select current and three others
            perm = draw(3, Npop - 1)
            perm[perm >= qq] += 1
            xi = pop[qq]
            z, R1, R2 = [pop[i] for i in perm]

            # Find the step direction and scale it to the length of the
            # projection of R1-R2 onto the step direction.
            # TODO: population sometimes not unique!
            step = xi - z
            denom = sum(step**2)
            if denom == 0:  # identical points; should be extremely rare
                step = EPS * rng.randn(*step.shape)
                denom = sum(step**2)
            step_scale = sum((R1 - R2) * step) / denom

            # Step using gamma of 2.38/sqrt(2) + U(-0.5, 0.5)
            gamma = 1.2 + rng.rand()
            delta_x[qq] = gamma * step_scale * step

            # Scale Metropolis probability by (||xi* - z||/||xi - z||)^(d-1)
            step_alpha[qq] = (norm(delta_x[qq] + step) / norm(step)) ** ((Nvar - 1) / 2)
            CR_used[qq] = 0.0

        elif alg[qq] == _DIRECT:  # Use one pair and all dimensions
            # Note that there is no F scaling, dimension selection or noise
            perm = draw(2, Npop - 1)
            perm[perm >= qq] += 1
            delta_x[qq, :] = pop[perm[0], :] - pop[perm[1], :]
            CR_used[qq] = 0.0

        else:
            raise RuntimeError("Select failed...should never happen")

        # Didn't implement this in the compiled version (yet)
        ## If no step was specified (exceedingly unlikely!), then
        ## select a delta at random from a gaussian approximation to the
        ## current population
        # if all(delta_x[qq] == 0):
        #    from numpy.linalg import cholesky, LinAlgError
        #    try:
        #        #print "No step"
        #        # Compute the Cholesky Decomposition of x_old
        #        R = (2.38/sqrt(Nvar)) * cholesky(cov(pop.T) + EPS*eye(Nvar))
        #        # Generate jump using multinormal distribution
        #        delta_x[qq] = dot(rng.randn(*(1, Nvar)), R)
        #    except LinAlgError:
        #        print("Bad cholesky")
        #        delta_x[qq] = rng.randn(Nvar)

    # Update x_old with delta_x and noise
    delta_x *= scale
    # print("alg", alg)
    # print("CR_used", CR_used)
    # print("delta_x", delta_x)

    # [PAK] The noise term needs to depend on the fitting range
    # of the parameter rather than using a fixed noise value for all
    # parameters.  The  current parameter value is a pretty good proxy
    # in most cases (i.e., relative noise), but it breaks down if the
    # parameter is zero, or if the range is something like 1 +/- eps.

    # absolute noise
    # x_new = pop[:Nchain] + delta_x + scale*noise*rng.randn(Nchain, Nvar)

    # relative noise
    x_new = pop[:Nchain] * (1 + scale * noise * rng.randn(Nchain, Nvar)) + delta_x

    # no noise
    # x_new = pop[:Nchain] + delta_x

    return x_new, step_alpha, CR_used


def _check():
    from numpy import arange, vstack, ascontiguousarray

    max_pairs, snooker_rate, eps, noise, scale = 2, 0.1, 0.05, 1e-6, 1.0
    nchain, npop, nvar, ncr = 4, 10, 7, 4

    pop = 100 * arange(npop * nvar, dtype="d").reshape((npop, nvar))
    pop = pop * (1 + rng.rand(*pop.shape) * 0.1)
    ratios = 1.0 / (rng.randint(4, size=ncr) + 1)  # 4 => ratios in [0.2, 0.25, 0.333, 0.5, 1.0]
    weights = [1 / ncr] * ncr  # equal-weight for each CR
    # print(ratios, weights)
    CR = ascontiguousarray(vstack((ratios, weights)).T, dtype="d")
    work = lambda: de_step(nchain, pop, CR, max_pairs=max_pairs, eps=eps)
    x_new, _step_alpha, used = work()
    # print(f"{pop=}")
    # print(f"{x_new=}")
    # print(f"{_step_alpha=}")
    # print(f"{used=}")
    print("""\
The following table shows the expected portion of the dimensions that
are changed and the rounded value of the change for each point in the
population.
""")
    for k, u in enumerate(used):
        rstr = f"{int(u * 100):4d}% " if u else " full "
        vstr = " ".join(f"{v:.2f}" for v in x_new[k] - pop[k])
        print(rstr + vstr)

    if 1:  # timing check
        from timeit import timeit
        from ctypes import c_double
        from .compiled import dll

        dll_work = lambda: dll.de_step(
            nchain,
            nvar,
            len(CR),
            pop.ctypes,
            CR.ctypes,
            max_pairs,
            c_double(eps),
            c_double(snooker_rate),
            c_double(noise),
            c_double(scale),
            x_new.ctypes,
            _step_alpha.ctypes,
            used.ctypes,
        )

        print("small pop time (ms)", timeit(work, number=10000) / 10)
        if dll:
            print("small pop time compiled (ms)", timeit(dll_work, number=10000) / 10)
        else:
            print("no dlls")

        nchain, nvar = 1000, 50
        npop = nchain
        pop = 100 * arange(npop * nvar, dtype="d").reshape((npop, nvar))
        pop = pop * (1 + rng.rand(*pop.shape) * 0.1)
        print("large pop time (ms)", timeit(work, number=100))
        if dll:
            x_new, _step_alpha, used = work()  # need this line to define new return vectors for dll
            print("large pop time compiled (ms)", timeit(dll_work, number=100))


if __name__ == "__main__":
    _check()



"""
Wx-Pylab magic for displaying plots within an application's window.
"""

from math import log10, floor
import string

import wx

import numpy as np


class EmbeddedPylab(object):
    """
    Define a 'with' context manager that lets you use pylab commands to
    plot on an embedded canvas.  This is useful for wrapping existing
    scripts in a GUI, and benefits from being more familiar than the
    underlying object oriented interface.

    As a convenience, the pylab module is returned on entry.

    Example
    -------

    The following example shows how to use the WxAgg backend in a wx panel::

        from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
        from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg as Toolbar
        from matplotlib.figure import Figure

        class PlotPanel(wx.Panel):
            def __init__(self, *args, **kw):
                wx.Panel.__init__(self, *args, **kw)

                figure = Figure(figsize=(1,1), dpi=72)
                canvas = FigureCanvas(self, wx.ID_ANY, figure)
                self.pylab_interface = EmbeddedPylab(canvas)

                # Instantiate the matplotlib navigation toolbar and explicitly show it.
                mpl_toolbar = Toolbar(canvas)
                mpl_toolbar.Realize()

                # Create a vertical box sizer to manage the widgets in the main panel.
                sizer = wx.BoxSizer(wx.VERTICAL)
                sizer.Add(canvas, 1, wx.EXPAND|wx.LEFT|wx.RIGHT, border=0)
                sizer.Add(mpl_toolbar, 0, wx.EXPAND|wx.ALL, border=0)

                # Associate the sizer with its container.
                self.SetSizer(sizer)
                sizer.Fit(self)

            def plot(self, *args, **kw):
                with self.pylab_interface as pylab:
                    pylab.clf()
                    pylab.plot(*args, **kw)

    Similar patterns should work for the other backends.  Check the source code
    in matplotlib.backend_bases for examples showing how to use matplotlib
    with other GUI toolkits.
    """

    def __init__(self, canvas):
        # delay loading pylab until matplotlib.use() is called
        from matplotlib.backend_bases import FigureManagerBase

        self.fm = FigureManagerBase(canvas, -1)

    def __enter__(self):
        # delay loading pylab until matplotlib.use() is called
        import pylab
        from matplotlib._pylab_helpers import Gcf

        # Note: don't need to track and restore the current active since it
        # will automatically be restored when we pop the current figure.
        Gcf.set_active(self.fm)
        return pylab

    def __exit__(self, *args, **kw):
        # delay loading pylab until matplotlib.use() is called
        from matplotlib._pylab_helpers import Gcf

        if hasattr(Gcf, "_activeQue"):  # CRUFT: MPL < 3.3.1
            Gcf._activeQue = [f for f in Gcf._activeQue if f is not self.fm]
            try:
                del Gcf.figs[-1]
            except KeyError:
                pass
        else:
            Gcf.figs.pop(self.fm.num, None)


class Validator(wx.PyValidator):
    def __init__(self, flag):
        wx.PyValidator.__init__(self)
        self.flag = flag
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def Clone(self):
        return Validator(self.flag)

    def Validate(self, win):
        return True

    def TransferToWindow(self):
        return True

    def TransferFromWindow(self):
        return True

    def OnChar(self, evt):
        key = chr(evt.GetKeyCode())
        if self.flag == "no-alpha" and key in string.letters:
            return
        if self.flag == "no-digit" and key in string.digits:
            return
        evt.Skip()


def nice(v, digits=4):
    """Fix v to a value with a given number of digits of precision"""
    if v == 0.0 or not np.isfinite(v):
        return v
    else:
        sign = v / abs(v)
        place = floor(log10(abs(v)))
        scale = 10 ** (place - (digits - 1))
        return sign * floor(abs(v) / scale + 0.5) * scale



import argparse
import multiprocessing
import os
from pathlib import Path
import sysconfig

CC = sysconfig.get_config_vars().get("CC", os.environ.get("CC", "gcc"))


def compile_dll(nthreads=None, use_openmp=True, dry_run=False):
    num_cores = multiprocessing.cpu_count()
    if nthreads == 0:
        nthreads = num_cores

    os.chdir(Path(__file__).parent)
    openmp_flag = "-fopenmp" if use_openmp else ""
    flags = f"-I ./random123/include/ -O2 {openmp_flag} -shared -lm -o _compiled.so -fPIC -DMAX_THREADS={nthreads}"
    compile_command = f"{CC} compiled.c {flags}"
    if dry_run:
        print(f"Would compile: {compile_command}")
        return False

    if not (Path(__file__).parent / "random123" / "include" / "Random123").exists():
        print("""\
        Random123 not found in the expected location: clone it with the following command:
        git clone --branch v1.14.0 https://github.com/DEShawResearch/random123 bumps/dream/random123
        """)

    print(f"Compiling: {compile_command}")
    os.system(compile_command)

    return os.path.exists("_compiled.so")


def remove_dll():
    dll_path = Path(__file__).parent / "_compiled.so"
    if dll_path.exists():
        print(f"Removing {dll_path}")
        dll_path.unlink()
    else:
        print(f"{dll_path} does not exist")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--nthreads",
        type=int,
        default=64,
        help="Number of threads to compile for, use 0 for all available cores, default is 64",
    )
    parser.add_argument(
        "--no-openmp",
        action="store_true",
        help="Disable OpenMP (e.g. for clang, where it is not supported)",
    )
    parser.add_argument("--remove", action="store_true", help="Remove the compiled dll")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Print the compile command without running it",
    )
    args = parser.parse_args()
    if args.remove:
        remove_dll()
    else:
        compile_dll(nthreads=args.nthreads, use_openmp=not args.no_openmp, dry_run=args.dry_run)


if __name__ == "__main__":
    main()



# This program is in the public domain
# Author: Paul Kienzle
"""
Random walk functions.

:function:`walk` simulates a mean-reverting random walk.
"""

# This code was developed to test outlier detection

__all__ = ["walk"]

from numpy import asarray, ones_like, nan, isnan

from . import util


def walk(n=1000, mu=0, sigma=1, alpha=0.01, s0=nan):
    """
    Mean reverting random walk.

    Returns an array of n-1 steps in the following process::

        s[i] = s[i-1] + alpha*(mu-s[i-1]) + e[i]

    with e ~ N(0,sigma).

    The parameters are::

        *n* walk length
        *s0* starting value, defaults to N(mu,sigma)
        *mu* target mean, defaults to 0
        *sigma* volatility
        *alpha* in [0,1] reversion rate

    Use alpha=0 for a pure Gaussian random walk or alpha=1 independent
    samples about the mean.

    If *mu* is a vector, multiple streams are run in parallel.  In this
    case *s0*, *sigma* and *alpha* can either be scalars or vectors.

    If *mu* is an array, the target value is non-stationary, and the
    parameter *n* is ignored.

    Note: the default starting value should be selected from a distribution
    whose width depends on alpha.  N(mu,sigma) is too narrow.  This
    effect is illustrated in :function:`demo`, where the following choices
    of sigma and alpha give approximately the same histogram::

        sigma = [0.138, 0.31, 0.45, 0.85, 1]
        alpha = [0.01,  0.05, 0.1,  0.5,  1]
    """
    s0, mu, sigma, alpha = [asarray(v) for v in (s0, mu, sigma, alpha)]
    nchains = mu.shape[0] if mu.ndim > 0 else 1

    if mu.ndim < 2:
        if isnan(s0):
            s0 = mu + util.rng.randn(nchains) * sigma
        s = [s0 * ones_like(mu)]
        for i in range(n - 1):
            s.append(s[-1] + alpha * (mu - s[-1]) + sigma * util.rng.randn(nchains))
    elif mu.ndim == 2:
        if isnan(s0):
            s0 = mu[0] + util.rng.randn(nchains) * sigma
        s = [s0 * ones_like(mu[0])]
        for i in range(mu.shape[1]):
            s.append(s[-1] + alpha * (mu[i] - s[-1]) + sigma * util.rng.randn(nchains))
    else:
        raise ValueError("mu must be scalar, vector or 2D array")
    return asarray(s)


def demo():
    """
    Example showing the relationship between alpha and sigma in the random
    walk posterior distribution.

    The lag 1 autocorrelation coefficient R^2 is approximately 1-alpha.
    """
    from numpy import mean, std, sum
    import pylab
    from matplotlib.ticker import MaxNLocator

    pylab.seed(10)  # Pick a pretty starting point

    # Generate chains
    n = 5000
    mu = [0, 5, 10, 15, 20]
    sigma = [0.138, 0.31, 0.45, 0.85, 1]
    alpha = [0.01, 0.05, 0.1, 0.5, 1]
    chains = walk(n, mu=mu, sigma=sigma, alpha=alpha)

    # Compute lag 1 correlation coefficient
    m, s = mean(chains, axis=0), std(chains, ddof=1, axis=0)
    r2 = sum((chains[1:] - m) * (chains[:-1] - m), axis=0) / ((n - 2) * s**2)
    r2[abs(r2) < 0.01] = 0

    # Plot chains
    ax_data = pylab.axes([0.05, 0.05, 0.65, 0.9])  # x,y,w,h
    ax_data.plot(chains)
    textkw = dict(
        xytext=(30, 0), textcoords="offset points", verticalalignment="center", backgroundcolor=(0.8, 0.8, 0.8, 0.8)
    )
    label = r"$\ \alpha\,%.2f\ \ \sigma\,%.3f\ \ " r"R^2\,%.2f\ \ avg\,%.2f\ \ std\,%.2f\ $"
    for m, s, a, r2, em, es in zip(mu, sigma, alpha, r2, m, s):
        pylab.annotate(label % (a, s, r2, em - m, es), xy=(0, m), **textkw)

    # Plot histogram
    ax_hist = pylab.axes([0.75, 0.05, 0.2, 0.9], sharey=ax_data)
    ax_hist.hist(chains.flatten(), 100, orientation="horizontal")
    pylab.setp(ax_hist.get_yticklabels(), visible=False)
    ax_hist.xaxis.set_major_locator(MaxNLocator(3))

    pylab.show()


if __name__ == "__main__":
    demo()
